//
// escape.hh: escaping special characters for output
// Copyright (C) 2003, 2004, 2012 Akim Demaille <akim@epita.fr> and
//                                Benoit Perrot <benoit@lrde.epita.fr>
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// It is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Nolimips.  If not, see <http://www.gnu.org/licenses/>.
//
#ifndef MISC_ESCAPE_HH
# define MISC_ESCAPE_HH

# include <ios>
# include <iomanip>
# include <iostream>
# include <string>
# include <cctype>

class escape
{
public:
  escape (const std::string& s) : s_ (s)
  {}

  const std::string& s_;
  bool print_delim_;
};

inline std::ostream&
operator<<(std::ostream& o, const escape& e)
{
  /* For some reason, when we use std::locale on Mac OS X, the
     following exception may be thrown:

       terminate called after throwing an instance of 'std::runtime_error'
         what():  locale::facet::_S_create_c_locale name not valid

     A workaround is to unset the environment variable `LANG' prior to
     using std::locale, but this is cumbersome.

     So we use the C locale-agnostic version of isprint() from header
     `cctype', instead of C++ isprint() from header `locale' taking a
     locale as second argument.

     See also
     http://stackoverflow.com/questions/1745045/stdlocale-breakage-on-macos-10-6-with-lang-en-us-utf-8
     on this topic.  */
  std::ios_base::fmtflags flags = o.flags (std::ios_base::hex);
  for (std::string::const_iterator p = e.s_.begin (); p != e.s_.end (); ++p)
    switch (*p)
      {
      case '\a': o << "\\a"; break;
      case '\b': o << "\\b"; break;
      case '\f': o << "\\f"; break;
      case '\n': o << "\\n"; break;
      case '\r': o << "\\r"; break;
      case '\t': o << "\\t"; break;

      case '\\': o << "\\\\"; break;
      case '\"': o << "\\\""; break;
      default:
        if (std::isprint (*p))
          o << *p;
        else
          o << "\\x"
            << std::setfill ('0') << std::setw (2) << (int) (unsigned char) *p;
      }
  o.flags (flags);
  return o;
}

#endif // !MISC_ESCAPE_HH
