//
// This file is part of Nolimips, a MIPS simulator with unlimited registers
// Copyright (C) 2003, 2004, 2006, 2009 Benoit Perrot <benoit@lrde.epita.fr>
//
// Nolimips is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// Nolimips is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Nolimips.  If not, see <http://www.gnu.org/licenses/>.
//

#include "misc/readline.hh"
#include "parse-tasks.hh"

#include "misc/lexutils.hh"

#include "parse/libparse.hh"

#include "inst/program.hh"
#include "inst/program_solver.hh"

#include "shell/shell.hh"
#include "shell/cmd.hh"

/* This standard is included _after_ local headers on purpose, to
   circumvent incompatibilities between C and C++ headers regarding
   the C macro / C++ function `isxdigit' (since g++ 4.3).  See e.g.
   http://gcc.gnu.org/ml/libstdc++/2003-01/msg00209.html for more
   information.  */
#include <iomanip>

namespace shell
{

  Shell &
  Shell::instance()
  {
    static Shell sh;
    return sh;
  }

#ifdef HAVE_READLINE_READLINE_H
  // --------------------------------------------------------------------------
  // Completion engine

  static std::string
  find_common_prefix(const std::vector<std::string> &l)
  {
    if (0 < l.size())
      {
        unsigned nb_common_chars = l[0].size();

        for (std::vector<std::string>::const_iterator
               e = l.begin(); e != l.end() && 0 < nb_common_chars; ++e)
          {
            unsigned c = 0;
            while (c < nb_common_chars && l.front()[c] == (*e)[c])
              ++c;
            nb_common_chars = c;
          }

        return std::string(l.front(), 0, nb_common_chars);
      }

    return std::string();
  }

  static char **
  shell_completer(const char *text, int /* start */, int end)
  {
//     if (rl_line_buffer[rl_point] && !isblank(rl_line_buffer[rl_point]))
//       return 0;

    // Generate possible completions
    std::string line_to_complete(rl_line_buffer, end);
    std::vector<std::string> completions;
    delete Shell::instance().parse(line_to_complete,
                                   &completions); // wonderfull

    char **matches = 0;
    if (0 < completions.size())
      {
        matches = (char**) malloc(sizeof(char*) * (completions.size() + 2));
        if (matches)
          {
            // Extract max common characters
            std::string substitution = find_common_prefix(completions);

            // Fill matches array
            if (!substitution.empty())
              matches[0] = strdup(substitution.c_str());
            else
              matches[0] = strdup(text);
            ++matches;
            for (unsigned c = 0; c < completions.size(); ++c)
              matches[c] = strdup(completions[c].c_str());
            matches[completions.size()] = 0;
            --matches;
          }
      }

    return matches;
  }

#endif // !HAVE_READLINE_READLINE_H

  // --------------------------------------------------------------------------
  // Shell
  // --------------------------------------------------------------------------

  Shell::Shell():
    program_(0)
  {
#ifdef HAVE_READLINE_READLINE_H
    rl_attempted_completion_function = shell_completer;
#endif // !HAVE_READLINE_READLINE_H

    vm_.get_cpu().set_check_callee_save(true);
    vm_.get_cpu().set_trace(false);

    map_token_["run"] = Cmd::id_run;
    map_token_["quit"] = Cmd::id_quit;
    map_token_["break"] = Cmd::id_break;
    map_token_["continue"] = Cmd::id_continue;
    map_token_["display"] = Cmd::id_display;
    map_token_["dump"] = Cmd::id_dump;
    map_token_["load"] = Cmd::id_load;
    map_token_["mem"] = Cmd::id_mem;
    map_token_["backtrace"] = Cmd::id_backtrace;
    map_token_["bt"] = Cmd::id_backtrace;
    map_token_["next"] = Cmd::id_next;
    map_token_["print"] = Cmd::id_print;
    map_token_["undisplay"] = Cmd::id_undisplay;
    map_token_["nop-after-branch"] = Cmd::id_nop_after_branch;
    map_token_["help"] = Cmd::id_help;

    map_register_["zero"] = vm::Cpu::zero;
    map_register_["at"] = vm::Cpu::at;
    map_register_["v0"] = vm::Cpu::v0;
    map_register_["v1"] = vm::Cpu::v1;
    map_register_["a0"] = vm::Cpu::a0;
    map_register_["a1"] = vm::Cpu::a1;
    map_register_["a2"] = vm::Cpu::a2;
    map_register_["a3"] = vm::Cpu::a3;
    map_register_["t0"] = vm::Cpu::t0;
    map_register_["t1"] = vm::Cpu::t1;
    map_register_["t2"] = vm::Cpu::t2;
    map_register_["t3"] = vm::Cpu::t3;
    map_register_["t4"] = vm::Cpu::t4;
    map_register_["t5"] = vm::Cpu::t5;
    map_register_["t6"] = vm::Cpu::t6;
    map_register_["t7"] = vm::Cpu::t7;
    map_register_["s0"] = vm::Cpu::s0;
    map_register_["s1"] = vm::Cpu::s1;
    map_register_["s2"] = vm::Cpu::s2;
    map_register_["s3"] = vm::Cpu::s3;
    map_register_["s4"] = vm::Cpu::s4;
    map_register_["s5"] = vm::Cpu::s5;
    map_register_["s6"] = vm::Cpu::s6;
    map_register_["s7"] = vm::Cpu::s7;
    map_register_["t8"] = vm::Cpu::t8;
    map_register_["t9"] = vm::Cpu::t9;
    map_register_["k0"] = vm::Cpu::k0;
    map_register_["k1"] = vm::Cpu::k1;
    map_register_["gp"] = vm::Cpu::gp;
    map_register_["sp"] = vm::Cpu::sp;
    map_register_["fp"] = vm::Cpu::fp;
    map_register_["ra"] = vm::Cpu::ra;
  }

  Shell::~Shell()
  {
    delete program_;
  }


  void
  Shell::run()
  {
    Cmd *command = new Cmd(Cmd::id_null);

    std::string line;
    while (execute(*command) == 0 && misc::readline("(nolimips) ", line))
      {
        display();

        if (!line.empty())
          {
            delete command;
            command = parse(line);
          }
      }

    delete command;
  }

  std::string
  Shell::eat_word(std::istringstream &iss)
  {
    misc::skip_blank(iss);

    std::string res;
    while (!isblank(iss.peek()))
      {
        char c = static_cast<char>(iss.get());
        if (iss.eof())
          break;
        res += c;
      }
    return res;
  }

  bool
  Shell::eat_register(std::istringstream &iss, std::string &reg)
  {
    reg = eat_word(iss);

    if (reg == "pc")
      return true;
    map_register_type::const_iterator it = map_register_.find(reg);
    if (it != map_register_.end())
      return true;
    // TODO: check unlimited register here
    return false;
  }

  bool
  Shell::eat_label(std::istringstream &iss, std::string &label)
  {
    label = eat_word(iss);

    if (label.empty())
      return false;
    char c = label.c_str()[0];
    if (isdigit(c))
      return false;
    return true;
  }

  Cmd::identifier_type
  Shell::eat_command_id(std::istringstream &iss,
                        std::vector<std::string> *completions)
  {
    Cmd::identifier_type id = Cmd::id_null;

    std::string word = eat_word(iss);
    bool complete_p = iss.eof() && completions;

    if (!word.empty())
      {
        map_token_type::const_iterator it = map_token_.lower_bound(word);
        if (it == map_token_.end() || it->first.find(word) != 0)
          id = Cmd::id_unknown;
        else
          {
            id = it->second;

            map_token_type::const_iterator next(it);
            while (next != map_token_.end() && next->first.find(word) == 0)
              {
                if (complete_p)
                  completions->push_back(next->first);

                if (it->first != next->first)
                  id = Cmd::id_ambiguous;
                ++next;
              }

            return id;
          }
      }

    if (complete_p)
      for (map_token_type::const_iterator
             it = map_token_.begin(); it != map_token_.end(); ++it)
        completions->push_back(it->first);

    return id;
  }

  Cmd*
  Shell::parse(const std::string &str,
               std::vector<std::string> *completions)
  {
    std::istringstream iss(str);
    Cmd* command = 0;

    Cmd::identifier_type id = eat_command_id(iss, completions);
    switch (id)
      {
      case Cmd::id_run:
        {
          command = new Cmd(Cmd::id_run);
          std::string tmp = eat_word(iss);
          while (0 < tmp.length())
            {
              command->push_arg(tmp);
              tmp = eat_word(iss);
            }
          return command;
        }

      case Cmd::id_break:
        {
          command = new Cmd(Cmd::id_break);
          std::string tmp;
          if (eat_label(iss, tmp))
            {
              command->push_arg(tmp);
              return command;
            }
          delete command;
          break;
        }
      case Cmd::id_load:
        {
          command = new Cmd(Cmd::id_load);
          std::string tmp = eat_word(iss);
          if (!tmp.empty())
            {
              command->push_arg(tmp);
              return command;
            }
          delete command;
          break;
        }

      case Cmd::id_print:
      case Cmd::id_display:
      case Cmd::id_undisplay:
        {
          command = new Cmd(id);
          std::string tmp;
          if (eat_register(iss, tmp))
            {
              command->push_arg(tmp);
              return command;
            }
          delete command;
          break;
        }

      case Cmd::id_mem:
        {
          command = new Cmd(Cmd::id_mem);

           // FIXME: check tmp is a memory address: eat_memory_address()
          std::string tmp = eat_word(iss);
          if (!tmp.empty())
            {
              command->push_arg(tmp);
              return command;
            }
          delete command;
          break;
        }

      case Cmd::id_quit:
      case Cmd::id_continue:
      case Cmd::id_next:
      case Cmd::id_dump:
      case Cmd::id_backtrace:
      case Cmd::id_nop_after_branch:
      case Cmd::id_help:

      case Cmd::id_unknown:
      case Cmd::id_ambiguous:
      case Cmd::id_null:
        return new Cmd(id);
      }

    return new Cmd(Cmd::id_unknown);
  }


  // --------------------------------------------------------------------------
  // Execution engine

  void
  Shell::do_load(const Cmd &cmd)
  {
    // Reset status
    delete program_;
    exit_status = exit_success;

    //
    program_ = ::parse::parse(*cmd.get_args().begin());
    if (exit_status == exit_success)
      {
        // FIXME: as soon as possible, put here nop-after-branches filling

        inst::ProgramSolver program_solver;
        //      program_solver.set_max_callee_save(max_callee_save);
        //      program_solver.set_max_caller_save(max_caller_save);

        program_solver.solve(*program_);
        if (exit_status == exit_success)
          {
            vm_.load_program(*program_);
          }
        else
          std::cerr << "Solving failed, program not loaded." << std::endl;
      }
    else
      std::cerr << "Parsing failed, program not loaded." << std::endl;

    //
    if (exit_status != exit_success)
      {
        delete program_;
        program_ = 0;
      }
  }


  void
  Shell::do_run()
  {
    if (!program_)
      {
        std::cerr << "No program loaded." << std::endl;
        return;
      }

    if (vm_.get_status() != vm::VirtualMachine::halt)
      vm_.set_runnable();
    else
      {
        vm_.reset();
        vm_.set_runnable();
      }
    vm_.execute(false);
  }

  void
  Shell::do_continue()
  {
    if (!program_)
      {
        std::cerr << "No program loaded." << std::endl;
        return;
      }

    if (vm_.get_status() != vm::VirtualMachine::halt)
      {
        vm_.set_runnable();
        vm_.execute(false);
      }
    else
      std::cerr << "The program is not being run." << std::endl;
  }

  void
  Shell::do_next()
  {
    if (!program_)
      {
        std::cerr << "No program loaded." << std::endl;
        return;
      }

    if (vm_.get_status() != vm::VirtualMachine::halt)
      {
        vm_.set_run_next();
        vm_.execute(true);
      }
    else
      std::cerr << "The program is not being run." << std::endl;
  }

  void
  Shell::do_break(const Cmd &cmd)
  {
    if (!program_)
        std::cerr << "No program loaded." << std::endl;
    else
      {
        misc::unique_string label(*cmd.get_args().begin());

        if (!program_->text_section().has_label(label))
          std::cerr << "Label " << label
                    << " not found." << std::endl;
        else
          {
            int offset =
              program_->text_section().get_offset(label);

            std::cout << "Breakpoint " << label
                      << " at " << offset << " (+4)." << std::endl;
            vm_.add_breakpoint(offset + 4);
          }
      }
  }

  void
  Shell::do_display(const Cmd &cmd)
  {
    add_display(*cmd.get_args().begin());
  }

  void
  Shell::do_undisplay(const Cmd &cmd)
  {
    remove_display(*cmd.get_args().begin());
  }

  void
  Shell::do_dump()
  {
    std::cout << vm_ << std::endl;
  }

  void
  Shell::do_print(const Cmd &cmd)
  {
    std::cout
      << *cmd.get_args().begin() << " = "
      << get_register(*cmd.get_args().begin()) << std::endl;
  }

  void
  Shell::do_nop_after_branch()
  {
    parse::tasks::nop_after_branch();
  }

  void
  Shell::do_mem()
  {
    //       const std::string* addr = *(list_args_.begin());
    //std::cout << "addr: " << *addr << " = " << sh_.vm_.get_memory_offset(atoi(addr->c_str())) << std::endl;
  }

  void
  Shell::do_backtrace() const
  {
    const std::vector<vm::register_type> &call_stack = vm_.get_call_stack();

    if (call_stack.empty())
      std::cout << "No stack." << std::endl;
    else
      {
        unsigned level = 0;
        for (std::vector<vm::register_type>::const_reverse_iterator
               it = call_stack.rbegin(); it != call_stack.rend(); ++it)
          {
            std::cout
              << "#" << level << "  0x"
              << std::hex << std::setfill('0') << std::setw(8) << (*it)
              << std::endl;
            ++level;
          }
      }
   }

  void
  Shell::do_unknown() const
  {
    std::cerr << "Unknown command" << std::endl;
  }
  void
  Shell::do_ambiguous() const
  {
    std::cerr << "Ambiguous command" << std::endl;
  }

  void
  Shell::do_help() const
  {
    std::cout
      << "run <args>" << std::endl
      << "\trun the current loaded file with the specified arguments"
      << std::endl
      << "load <filename>" << std::endl
      << "\tload the specified filename" << std::endl
      << "break <label>" << std::endl
      << "\tenable a break point on a label" << std::endl
      << "quit" << std::endl
      << "\tquit the debugging session" << std::endl
      << "next" << std::endl
      << "\texecute the next asm line" << std::endl
      << "continue" << std::endl
      << "\tcontinue to the next breakpoint" << std::endl
      << "display <register>" << std::endl
      << "\tdisplay (and set displayable) a register value"
      << std::endl
      << "undisplay <register>" << std::endl
      << "\tundisplay a register" << std::endl
      << "print <register>" << std::endl
      << "\tprint a register value" << std::endl
      << "dump" << std::endl
      << "\tdump all the register value" << std::endl
      << "mem <addr>" << std::endl
      << "\tprint a the value of a memory region" << std::endl
      << "backtrace" << std::endl
      << "\tPrint backtrace of stack frame" << std::endl
      << "nop-after-branch" << std::endl
      << "\tenable Fill delay slot of branch instructions "
      << "with NOP" << std::endl;
  }

  /// Execute the Cmd.
  int
  Shell::execute(const Cmd &cmd)
  {
    switch (cmd.get_identifier())
      {
      case Cmd::id_quit:
        return 1;
      case Cmd::id_unknown:
        do_unknown();
        break;
      case Cmd::id_ambiguous:
        do_ambiguous();
        break;
      case Cmd::id_load:
        do_load(cmd);
        break;
      case Cmd::id_run:
        do_run();
        break;
      case Cmd::id_continue:
        do_continue();
        break;
      case Cmd::id_next:
        do_next();
        break;
      case Cmd::id_dump:
        do_dump();
        break;
      case Cmd::id_print:
        do_print(cmd);
        break;
      case Cmd::id_display:
        do_display(cmd);
        break;
      case Cmd::id_undisplay:
        do_undisplay(cmd);
        break;
      case Cmd::id_mem:
        do_mem();
        break;
      case Cmd::id_backtrace:
        do_backtrace();
        break;
      case Cmd::id_break:
        do_break(cmd);
        break;
      case Cmd::id_nop_after_branch:
        do_nop_after_branch();
        break;
      case Cmd::id_null:
        break;
      case Cmd::id_help:
        do_help();
      }
    return 0;
  }

} // namespace shell
